#ifndef _CONSTRAINT_H_
#define _CONSTRAINT_H_

#include "Common.h"
#include "StorableObject.h"

const int MIN_CONSTRAINT_ID = 600;
const int UNDEF_CONSTRAINT_ID = 600;

/**
  Eine Nebenbedingung zu einem Wunsch. Eine Nebenbedingung hat zwar wie ein
  Wunsch einen Nutzen, kann aber keinen konkreten Handlungsplan erzeugen.

  In unserem Modell ist die Nuetzlichkeit einer Nebenbedingung nur von der
  Situation, nicht aber von dem Wunsch, bei dessen Umsetzung sie erfuellt wird,
  abhaengig. (Wohl aber kann es verschiedene Wuensche verschieden viel kosten,
  dieselbe Nebenbedingung zu erfuellen).

  Auch von Constraint selbst koennen Instanzen erzeugt werden. Sie werden zur
  Initialisierung im Deliberator verwendet. Normalerweise werden aber nur von
  Nachkommen von Constraint Instanzen erzeugt.

  {\small {\bf Quelltext}

  [\URL[Header]{src/Constraint.h} | [\URL[Implementierung]{src/Constraint.cc}]

  {\bf CVS Informationen zum Interface}
  \begin{verbatim}
  **************************************************

  *
  * Aktueller Stand
  *
  *   $RCSfile: Constraint.h,v $
  *   $Revision: 1.12 $
  *   $Date: 1998/08/14 13:20:09 $
  *   $Author: myritz $ 
  *
  **************************************************
  \end{verbatim} }*/
class Constraint : public StorableObject
{
public:
  /** Liefert eine triviale Nebenbedingung, deren Erfuellung nichts bringt. */
  Constraint();
  
  /** Liefert den erwarteten Nutzen. Diese Methode ist const, obwohl sie
      interne Werte umsetzt (simStepOfUtilityCalculation und 
      theExpectedUtility).
      @return Ein Wert zwischen MIN_UTILITY und MAX_UTILITY. */
  int expectedUtility() const;
  
  /** Liefert den Typ der Nebenbedingung zurueck. */
  int id() const;
 
   /** Legt eine Kopie des Objekts auf dem Heap an und liefert einen Zeiger
      darauf zurueck. */
  virtual StorableObject* copy() const;
  
  /** Beschreibung als String fuer die Ausgabe auf einem Terminal. */
  virtual String toString() const;

  /** Virtueller Destruktor */
  virtual ~Constraint();

protected:

  // Variablen
  /** Typ des Wunsches. Wird sofort im Konstruktor gesetzt. */
  int theId;  

  /** Hier wird die ExpectedUtility gecacht. */
  mutable int theExpectedUtility;
  
  /** Ist der simStep im Weltmodell verschieden von diesem Wert, wird die
      ExpectedUtility neu berechnet. */
  mutable int simStepOfUtilityCalculation;


  // Methoden

  /** Fuehrt die eigentliche Berechnung der expectedUtility durch.
      @return Im Allgemeinen ein Wert zwischen zwischen NEUTRAL_UTILITY und
      MAX_UTILITY, bei Constraint selbst aber NEUTRAL_UTILITY. */
  virtual int computeExpectedUtility() const;

};

inline int Constraint::id() const
{
  return theId;
}

#endif

