#ifndef _FIELD_H_
#define _FIELD_H_

#include "Common.h"
#include "Array.h"
#include "Vector.h"
#include "ViewPoint.h"
#include "StableObject.h"
#include "Exception.h"

/**
  Kapselung des Wissens ueber das Layout des Spielfeldes. Hier koennen absolute
  Positionen von Flaggen, Toren und Linien erfragt werden. Ausserdem findet
  sich hier die Berechnung eines Blickpunktes aus einer
  Reihe von relativen Positionsangaben. 

  Sinn der Klasse ist es, die Daten eines Fussballfeldes wegzukapseln.

  {\small {\bf Quelltext}

  [\URL[Header]{src/Field.h} | [\URL[Implementierung]{src/Field.cc}]

  {\bf CVS Informationen zum Interface}
  \begin{verbatim}
  **************************************************
  *
  * Aktueller Stand
  *
  *   $RCSfile: Field.h,v $
  *   $Revision: 1.21 $
  *   $Date: 1998/08/14 13:20:32 $
  *   $Author: myritz $ 
  *
  **************************************************
  \end{verbatim} }*/
class Field
{
public:

  /** Erzeugt ein neues Feld. Bei der Erzeugung wird die aktuelle Breite
    des Tores aus dem globalen Configuration-Objekt gelesen und die
    Positionen der entsprechenden Flaggen neu gesetzt. */
  Field();

  /** Standpunktbestimmung ist fehlgeschlagen. */
  class FailedTriangulationException: public Exception {
  public:
    FailedTriangulationException()
      : Exception() {}
  };

  /** Positionsbestimmung ist fehlgeschlagen. */
  class FailedPositionCalculationException: public Exception {
  public:
    FailedPositionCalculationException()
      : Exception() {}
  };
 
  /** 
    Berechnet den Blickpunkt. Enthaelt das Linien-Array nur ein Objekt, wird
    davon ausgegangen, dass man sich innerhalb des Spielfeldes befindet.
    @param seenLines Die gesehenen Linien. 
    @param seenFlags Die gesehenen Flaggen und Tore. 
    @return ViewPoint-Objekt, das den berechneten Blickpunkt enthaelt.
    @exception FailedTriangulationException Kann auftreten, wenn entweder
    die Ids der uebergebenen Objekte nicht vom erwarteten Typ sind (also der
    erste Parameter keine Linie oder der zweite keine Flagge/Tor ist.
    Ausserdem wird die Exception erzeugt, wenn ein \"unmoeglicher" Fall beim
    Triangulieren auftritt.
    @exception FailedPositionCalculationException Wird erzeugt, wenn eine 
    Linie aber keine Flagge gesehen wurde. (Die Blickrichtung kann bestimmt
    werden, aber nicht die Position des Spielers.) */
  ViewPoint triangulateViewPoint( const Array<StableObject>& seenLines,
				  const Array<StableObject>& seenFlags ) const;
    
  /** Liefert YES, wenn sich der Punkt innerhalb des Spielfeldes befindet. */
  static bool isInsideField(const Vector& aPoint);
  
  /** Gibt die Position eines Punktes zurueck. Ein Punkt kann entweder eine
    Flagge oder ein Tor sein. Weil die Positionen einzelner Torflaggen vor
    dem Spiel noch nicht feststehen (Torbreite wird aus server.conf ausgelesen)
    ist diese Methode keine Klassenmethode.
    @param anId Die Id eines Punktes (wie in Gemeinsame Deklarationen)
    @return Vektor, der die absolute Position des Punktes auf dem Spielfeld
    beschreibt.
    @exception InvalidIdException wenn die Id keine
    existierende Flagge/Tor beschreibt. */
  const Vector& posOfPoint(int anId) const;
  
  /** Gibt die Laenge des Feldes zurueck.  */
  static double fieldLength();
  
  /** Gibt die Breite des Feldes zurueck.  */
  static double fieldWidth();

  /** Gibt die Breite des Tores zurueck, diese wird im Konstrukter aus
    dem Configuration-Objekt ausgelessen und damit aktualisiert. */
  double goalWidth() const;
  
  /** Gibt die Laenge des Strafraums zurueck.  */
  static double penaltyAreaLength();
  
  /** Gibt die Breite des Strafraums zurueck.  */
  static double penaltyAreaWidth();	
  
  /** Gibt den Radius des Anstosskreis zurueck. */
  static double kickoffCircleRadius();
  
  /** Berechnet die absolute Blickrichtung. Es wird eine der uebergebenen
    Linien ausgewaehlt, um daraus den Blickwinkel zu bestimmen. 
    @param seenLines Array mit den gesehenen Linien.
    @exception FailedTriangulationException Wird erzeugt, wenn keine Linie
    gesehen wurde, oder wenn sich aus der ausgewaehlten Linie der Blickwinkel
    nicht eindeutig ergibt. (Das ist in einem selten auftretenden Spezialfall
    fuer die rechte Linie der Fall.) */
  static double computeFaceDir( const Array<StableObject>& seenLines );
  
private:
  
  /** Berechnet aus der relativen Position einer mit ihrer Id gegebenen Flage
    und dem absoluten Blickwinkel, die eigene absolute Position. */
  Vector computePosFromFlag( const StableObject& flag, double faceDir ) const;
  
  
  // jetzt die internen Klassenvariablen;
  
  /** C-Array der Positionen aller Flaggen. Als Index kann FlagId - MINFLAG 
    verwendet werden. Das Array hat FLAGCOUNT Eintraege. */
  static Vector theFlagPositions[];
  
  /** C-Array der Positionen der Tore. Als Index kann GoalId - MINGOAL 
    verwendet werden. Das Array hat GOALCOUNT Eintraege. */
  static const Vector theGoalPositions[];
  
  /** Die Feldlaenge */
  static const double theFieldLength;
  
  /** Die Feldbreite */
  static const double theFieldWidth;
  
  /** Die Torbreite, diese wird aus den Configuration-Objekt aktualisiert */
  static double theGoalWidth;
  
  /** Die Strafraumlaenge */
  static const double thePenaltyAreaLength;
  
  /** Die Strafraumbreite */
  static const double thePenaltyAreaWidth;
  
  /** Der Radius des Anstosskreises */
  static const double theKickoffCircleRadius;
};

// Die Implementierung der Inline-Funktionen

inline double Field::fieldLength()
{ 
  return theFieldLength;
}

inline double Field::fieldWidth()
{
  return theFieldWidth;
}

inline double Field::goalWidth() const
{
  return theGoalWidth;
}

inline double Field::penaltyAreaLength()
{
  return thePenaltyAreaLength;
}

inline double Field::penaltyAreaWidth()
{
  return thePenaltyAreaWidth;
}  

inline double Field::kickoffCircleRadius()
{
  return theKickoffCircleRadius;
}   
#endif



