#ifndef _PASSIVE_OPTION_H_
#define _PASSIVE_OPTION_H_

#include "Option.h"
#include "Configuration.h"
#include "WorldModel.h"
#include "Skills.h"


/**
  Diese Klasse ist Superklasse fuer alle passiven W\"unsche.
  
  {\small {\bf Quelltext}

  [\URL[Header]{src/PassiveOption.h} | [\URL[Implementierung]{src/PassiveOption.cc}]

  {\bf CVS Informationen zum Interface}
  \begin{verbatim}
  **************************************************
  *
  * Aktueller Stand
  *
  *   $RCSfile: PassiveOption.h,v $
  *   $Revision: 1.15 $
  *   $Date: 1998/08/14 13:21:05 $
  *   $Author: myritz $ 
  *
  **************************************************
  \end{verbatim} }*/
class PassiveOption : public Option
{
public:

  /** @name M\"ogliche Gr\"unde die einen Abbruch des laufenden Planes
    verursachen k\"onnen. */
  //@{
    ///
    const int CANCEL_REASON_REFEREE_DECISION = 1;
    ///
    const int CANCEL_REASON_NEW_BALL_INFO = 2;
    ///
    const int CANCEL_REASON_CONT_TIME_EXPIRED = 4;
    ///
    const int CANCEL_REASON_PLAN_COMPLETED = 8;
    ///
    const int ALL_CANCEL_REASONS = 15;
  //@}
  
  /** Gibt die Abbruchkosten an. 
    @return NEUTRAL_COST */
  virtual int cancellationCost() const;
  
  /** Prueft, ob die Absicht tatsaechlich umsetzbar ist.
    @return YES */
  virtual bool isPossible() const;  
 
  /** Soll die Absicht unbedingt beibehalten werden?
    Manche Optionen koennen oder sollen nicht jederzeit abgebrochen werden.
    Diese Methode wird vom Deliberator vor jedem Entscheidungsvorgang
    gefragt, ob die aktuelle Absicht weitergemacht werden soll.
    Falls die \Ref{forceContinuationTime} noch nicht erreicht wurde und die
    Daten des Weltmodells keinen aktuellen Ball enthalten wird die
    Absicht weiter verfolgt, also YES zur\"uckgegeben, sonst NO.
    @return YES oder NO. */
  virtual bool forceContinuation() const;

protected:
  /** Protected Konstruktor macht die Klasse abstrakt.
    Erzeugt eine PassiveOption. */
  PassiveOption();

  /** Der Grundnutzen des Wunsches.
    Diese Funktion testet, ob die Grundvoraussetzungen f\"ur eine Entscheidung 
    f\"ur einen passiven Wunsch vorliegen, d.h. die eigene Position ist
    ausreichend zuverl\"assig und der Ball ist entweder unzuverl\"assig, d.h.
    er wurde noch nie gesehen, oder ist au\3erhalb des KickRanges.
    @return NEUTRAL_UTILITY, wenn die Positionen nicht ausreichend genau sind;
    MIN_UTILITY, wenn der Ball innerhalb der KickRange ist; MAX_UTILITY, wenn
    sich eine passive Entscheidung anbietet. */
  virtual int baseUtility() const;

  /** Liefert die Kosten, die durch die Einhaltung einer Nebenbedingung   
    entstehen. Viele passive Optionen haben mit laufen zu tun, deshalb
    k\"onnen hier schon die Kosten f\"ur einige Constraints in Abh\"angigkeit
    vom Zielpunkt ermittelt werden.
    @param constraint Zeiger auf die einzuhaltende Nebenbedingung
    @return Im allgemeinen Fall ein Wert zwischen NEUTRAL_COST und MAX_COST. */
  virtual int constraintCost( const Constraint *constraint ) const;

  /** Setzt die Gr\"unde, die wenn sie eintreffen, zum Abbruch der Option
    f\"uhren. 
    @param reasons Eine Menge von Gr\"unden, die mittels 'bin\"aren oder'
    verkn\"unft werden sollen. */
  void setCancelReasons( int reasons = ALL_CANCEL_REASONS ) const;
  
  /** Entfernt Gr\"unde, die zum Abbruch f\"uhren k\"onnen.
    @param reasons Eine Menge von Gr\"unden, die mittels 'bin\"aren oder'
    verkn\"unft werden sollen. */
  void noCancelByReasons( int reasons ) const;

  /** Gibt die Gr\"unde zur\"uck die zum Abbruch der Option f\"uhren k\"onnen.
    @return Eine Menge von Gr\"unden, die mittels 'bin\"aren und'
    abgefragt werden sollen. */
  int cancelReasons() const;

  /** Gibt YES zur\"uck, wenn der angebene Abbruchsgrund als aktiv gilt. */
  bool isCancelReasonActive( int aReason ) const;

  /** Konstante fuer setForceContinuationTime().
    Wird angegeben, wenn die Option nicht an ihrer Fortsetzung interessiert ist. */
  static const int NO_CONTINUATION = -1;
  
  /** Setzt die Zeit bis zu der die passive Option ohne Bedenken weiter
    verfolgt werden kann. */ 
  void setForceContinuationTime( int timeSteps ) const;

  /** Erfragt die Zeit bis zu der die passive Option ohne Bedenken weiter
    verfolgt werden kann. */ 
  int forceContinuationTime() const;
  
  /** Liefert Ziel, das nicht im Abseits liegt
    @return Globale Position des Ziels */
  Vector closestNonOffsideTarget( const Vector& aTarget ) const;

  /** Setzt das passive Ziel. */
  void setTarget( const Vector& aTarget ) const;

  /** Erfragt das passive Ziel. */
  const Vector& target() const;

  /** Setzt die Variable die aussagt, ob das Ziel schon berechnet wurde.
   Mit dieser Methode wird auch gleich die Ausf\"uhrbarkeit der Option
   mitgesetzt (siehe \Ref{setIsOptionPossible}). */
  void setIsTargetComputed( bool isTargetComputed ) const;

  /** Erfragt die Variable die aussagt, ob das Ziel schon berechnet wurde. */
  bool isTargetComputed() const;

  /** Setzt die Variable die ansagt, da\3 der Plan abgearbeitet wurde. */
  void setIsPlanCompleted( bool isPlanCompleted = YES ) const;

  /** Erfragt, ob der Plan bereits abgearbeitet wurde. */
  bool isPlanCompleted() const;
  
  /** Erzeugt einen Plan, der den Agenten auf die gew\"unschte Position bringt.
    Wenn m\"oglich, wird gebeamt, sonst mu\3 wird gotoRegion gemacht. Dabei
    werden Hindernisse immer umgangen.
    @param target Absoluter Zielvektor
    @param targetRadius Zielradius f\"ur gotoRegion. Kann gebeamt werden, wird
    genau auf die gew\"unschte Position gebeamt.
    @param staminaLimit fuer gotoRegion. 
    @see gotoRegion
    @see beamToPosition */
  Plan movementTo( 
      const Vector& target, 
      const double targetRadius = TRIANGULATION_ACCURACY, 
      const double staminaLimit = 0,
      const int specialBehaviors = Skills::AVOID_OBSTACLES ) const;

  /** Erzeugt einen Plan, der wenn es sinnvoll ist etwas
    zu sagen, eine say-Action enth\"alt die die Daten des ganzen 
    Weltmodels enth\"alt. */
  Plan isSomethingToSay() const;

  /** Testet die \"ubergebene Position darauf, ob sie innnerhalb der
    goalieDangerousZone liegt. */
  bool isDangerousPos( const Vector& aPos ) const;
  
  /** Checkt, ob eine zu verteidigende Situation ( Freisto\"s, Einwurf, Ecke
    des Gegners ) vorliegt. */
  bool isDefenseSituation() const;

  /** Beschreibung als String fuer die Ausgabe auf einem Terminal. 
    Diese Methode gibt die Daten der Option aus, ohne den Optionnamen. */
  virtual String toString() const;
  
private:

  /** Eine Menge von Gr\"unden, die mittels bin\"arer Operationen gesetzt
    und abgefragt werden sollen. */
  mutable int theCancelReasons;

  /** Zeit bis zu der die passive Option ohne Bedenken weiter verfolgt
    werden kann. */
  mutable int theForceContinuationTime;

  /** Das passive Ziel. */
  mutable Vector theTarget;
  
  /** Variable die aussagt, ob das Ziel schon berechnet wurde. */
  mutable bool theIsTargetComputed;

  /** Variable die aussagt, ob der Plan bereits abgearbeitet wurde. */
  mutable bool theIsPlanCompleted;

};

inline void PassiveOption::setCancelReasons( int reasons ) const
{
  theCancelReasons = reasons;
  if( ! ( theCancelReasons & ALL_CANCEL_REASONS ) ) {
    throwException( Exception( "No cancel reason for the plan given." ) );
  }
}

inline void PassiveOption::noCancelByReasons( int reasons ) const
{
  theCancelReasons = theCancelReasons & ( ~reasons ); 
  if( ! ( theCancelReasons & ALL_CANCEL_REASONS ) ) {
    throwException( Exception( "No cancel reason for the plan left." ) );
  }
}

inline int PassiveOption::cancelReasons() const
{
  return theCancelReasons;
}

inline bool PassiveOption::isCancelReasonActive( int aReason ) const
{
  if( theCancelReasons & aReason ) {
    return YES;
  }
  else {
    return NO;
  }
}

inline void PassiveOption::setForceContinuationTime( int timeSteps ) const
{
  theForceContinuationTime = globalWorldModel->simStep() +
    min( globalConf->maxForceContinuationTime, timeSteps ) - 1;
}

inline int PassiveOption::forceContinuationTime() const
{
  return theForceContinuationTime;
}

inline void PassiveOption::setTarget( const Vector& aTarget ) const
{
  theTarget = aTarget;
  setIsTargetComputed( YES );
}

inline const Vector& PassiveOption::target() const
{
  return theTarget;
}

inline void PassiveOption::setIsTargetComputed( bool isTargetComputed ) const
{
  theIsTargetComputed = isTargetComputed;
  setIsOptionPossible( isTargetComputed );
}

inline bool PassiveOption::isTargetComputed() const
{
  return theIsTargetComputed;
}

inline void PassiveOption::setIsPlanCompleted( bool isPlanCompleted ) const
{
  theIsPlanCompleted = isPlanCompleted;
}

inline bool PassiveOption::isPlanCompleted() const
{
  return theIsPlanCompleted;
}

#endif



