#ifndef _STRING_H_
#define _STRING_H_

#include <string.h>

#include "Common.h"
#include "Exception.h"
#include "StorableObject.h"

/**
  Diese Klasse verwaltet einen String. 

  
  {\small {\bf Quelltext}
  
  [\URL[Header]{src/String.h}] | [\URL[Implementierung]{src/String.cc}] 
      
  {\bf CVS Informationen zum Interface}   
  \begin{verbatim} 
  **************************************************
  *
  * Aktueller Stand
  *
  *   $RCSfile: String.h,v $
  *   $Revision: 1.26 $
  *   $Date: 1998/08/14 13:21:24 $
  *   $Author: myritz $ 
  *
  **************************************************
  \end{verbatim} } */
#include <iostream.h>


class String : public StorableObject
{
public:
  /** @name Fehlerklassen */
  //@{
    /** Allgemeine Fehlerklasse */ 
    class StringException: public Exception {
    public:
      StringException(const String& aDescription )
        : Exception( aDescription ) {}
    };

    /** Umwandlungsfehlerklasse. 
      Sie wird bei fehlschlagenden Umwandlungen, z.b. in einen int-Wert 
      erzeugt. */
    class ConversionException: public StringException {
    public:
      ConversionException(const String& aDescription )
        : StringException( aDescription ) {}
    };
    
  //@}  
  
  /** @name Strings anlegen und zerst\"oren */
  //@{
    /** Neuer String aus einer Zeichenkette.
        @param newValue Der String wird mit dieser Zeichenkette initialisiert. */
    String (const char* newValue="");

    /** Neuer String aus einem int-Wert.
      @param newValue Der String wird mit einer Zeichenkette, die diesen
      int-Wert darstellt, initialisiert. */
    String (const int newValue);

    /** Neuer String aus einem long-Wert.
      @param newValue Der String wird mit einer Zeichenkette, die diesen
        long-Wert darstellt, initialisiert. */
    String (const long newValue);

    /** Neuer String aus einem double-Wert.
      @param newValue Der String wird mit einer Zeichenkette, die diesen
        double-Wert darstellt, initialisiert. */
    String (const double newValue);

    /** Der Copy-Konstruktor
        @param s Der String wird mit den Werten des Strings s initialisiert. */
    String (const String& s);

    /** Legt eine Kopie des Objekts auf dem Heap an.
      @see StorableObject::copy
      @return Zeiger auf die neu allozierte Kopie */
    virtual StorableObject* copy() const;

    /** Destruktor.
      Der Belegte Speicher wird wieder freigegeben. */
    virtual ~String ();

    //@}

      
  /** @name Anfragen und Tests  */
  //@{  
    /** Test auf Gleichheit */
    friend bool operator== (const String&, const String&);
    
    /** Test auf Ungleichheit */
    friend bool operator!= (const String&, const String&);

    /** Test, ob der String mit dem angegebenen String anfaengt.
      @param string wird mit dem Anfang des Strings verglichen.
      @return YES, wenn der Anfang des Strings mit dem Argument uebereinstimmt,
        NO sonst. */
    bool beginsWith( const String& string ) const;

    /** Liefert die Laenge des Strings. Das abschlie\3ende 0-Byte wird nicht
      mitgez\"ahlt.
      @return Anzahl der Zeichen im String. */
    int length() const;
  //@}
    
  
  /** @name Inhalt \"andern */
  //@{
    /** Einen zweiten String ang\"angen */
    friend String operator+ (const String&, const String&);

    /** Zuweisung eines neuen Wertes. */
    const String& operator= (const String& s);
    
  //@}

        
  /** @name Teilstrings erfragen */
  //@{
    /** Liefert einen Teilstring, links beginnend.
      @param charCount Laenge des Teilstrings. Ist die Laenge kleiner als 0,
        wird ein leerer String zurueckgegeben. Ist die Laenge groesser als 
        die Stringlaenge, wird der gesamte String zurueckgegeben.
      @return Anfangsstueck des Strings mit der gewaehlten Laenge */
    String leftString( int charCount ) const;
    /** Liefert einen Teilstring, rechts endend.
      @param charCount Laenge des Teilstrings. Ist die Laenge kleiner als 0,
        wird ein leerer String zurueckgegeben. Ist die Laenge groesser als 
        die Stringlaenge, wird der gesamte String zurueckgegeben.
      @return Endstueck des Strings mit der gewaehlten Laenge */
    String rightString( int charCount ) const;
    /** Liefert einen Teilstring.
      @param leftIndex Index bei dem der Teilstring anfangen soll. Das erste
        Zeichen des Strings hat den Index 0. Ist der Index kleiner als 0, wird
        bei 0 begonnen. Ist der Index groesser als die Stringlaenge, dann wird
        ein leerer String zur\"uckgegeben.
      @param charCount Laenge des Teilstrings. Ist die Laenge kleiner als 0,
        wird ein leerer String zurueckgegeben. Falls die Laenge ab dem Index
        die Laenge des Strings ueberschreitet, wird sie so gesetzt, das sie bis
        zum Ende des Strings reicht.
      @return Teilstueck des Strings in der gewaehlten Laenge, beginnend ab
        dem gewaehlten Index. */
    String subString( int leftIndex, int charCount = UNDEF_INT ) const;
  //@}

  
  /** @name Konvertierungen */
  //@{
    /** cString (also Char-Ptr zurueckgeben). 
      Der String darf nicht ge\"andert werden. Der gelieferte Pointer ist nur
      solange g\"ultig, wie das String-Objekt existiert, da\3 ihn geliefert
      hat. 
      @return Char-Zeiger auf den String */
    const char *cString() const;

    /** Liefert einen Int, indem es den eigenen Wert als Dezimalzahl
      interpretiert. 
      
      Hier sollte eigentlich eine String::ConversionException erzeugt werden,
      falls die Konvertierung fehlschl\"agt. Wer es schafft, atoi den
      entsprechenden Fehlerstatus zu entlocken, darf mir gerne die neue Routine
      zukommen lassen. 
      @return der ensprechende int-Wert. Schl\"agt die Konvertierung fehl, wird
        0 geliefert. */
    int asInt() const;
    /** Beschreibung als String fuer die Ausgabe auf einem Terminal. String
    selbst liefert nur sich selbst zur\"uck, die allgemeine ( Foo ... )
    Konvention verletzend. 
      @return Der String selbst */
    virtual String toString() const;
  //@}
  


private:
  /// Char-Pointer, der auf den eigentlichen Inhalt zeigt.
  char *value;
  
  /** Laenge des Strings */
  int len;
  
  /** Konstruktor f\"ur uninitialisierten String der gegebenen L\"aenge.
    Der Konstruktor wird von \Ref{operator+} und \Ref{subString} verwendet.
      @param length Die gew\"unschte L\"ange
      @param dummyParameter, dessen Wert nicht interpretiert wird. Er ist
      notwendig, um den Konstruktor von String(int) zu unterscheiden. */
  String(int length, char dummyParameter);
};


inline const char *String::cString() const 
{
  return value;
}

inline int String::length() const 
{
  return len;
}

inline bool operator== (const String& s1, const String& s2)
{ 
  return ( ( s1.value == s2.value ) || strcmp( s1.value, s2.value ) == 0 );
}

inline bool operator!= (const String& s1, const String& s2)
{
  return !(s1==s2);
}

inline bool String::beginsWith( const String& string ) const 
{
  return (strncasecmp(cString(), string.cString(), string.length()) == 0);
}

#endif










