
/**
 * Aktueller Stand
 *
 *   $RCSfile: WorldModel.cc,v $
 *   $Revision: 1.93 $
 *   $Date: 1998/08/14 13:21:32 $
 *   $Author: myritz $ 
 *
 */

#include "WorldModel.h"
#include "Log.h"
#include "String.h"
#include "InvalidIdException.h"
#include <math.h>


WorldModel *globalWorldModel;

WorldModel::WorldModel(Sensors *sensors) : theSensors( sensors ) {
  theField = Field();
  theInitiative = NO_TEAM_ID;
  thePlayerNumber = UNDEF_INT;
  theSide = NO_SIDE_ID;
  historyCount = globalConf->historyCount;
  situationFieldSize = 2*historyCount + 1;
  theSituations = Array<Situation>( situationFieldSize );
  theSimStepOfLastCatch = 0;
  int i;
  for ( i = 0; i < situationFieldSize; i++ ) {
    theSituations.addObject( Situation( 0 ) );
  }
  currentSituationIndex = historyCount;
  ModelledBall ball = ModelledBall( BALL_ID, Vector(0, 0), Vector(0, 0),
                                    Reliability::PERFECT,
                                    Reliability::PERFECT );
  theSituations[currentSituationIndex].setBall( ball );

  theInitialPositions.addObject( Vector() );
  theInitialPositions.addObject( Vector() );
  theInitialPositions.addObject( Vector() );
  
  theRole = GENERIC_ROLE;

}

void WorldModel::setPlayerNumber( int newPlayerNumber )
{
  thePlayerNumber = newPlayerNumber;

  try {
    situation().playerWithId( id() );
  }
  catch (Situation::NoPlayerWithThisIdException) {
    ModelledPlayer playerSelf = ModelledPlayer( id(), Vector( -5.0, 37.0 ),
                                                Vector( 0.0, 0.0 ), 0.0 );
    playerSelf.setSpeedReliability( Reliability::PERFECT );
    playerSelf.setVisibleAngle( VIEW_WIDTH_NORMAL );
    playerSelf.setStaminaReliability( Reliability::PERFECT );
    theSituations[currentSituationIndex].addPlayer( playerSelf );
  }
  
  // Positionen und Rolle setzen
  setInitialPosition( BEFORE_KICK_OFF, OUR_TEAM_ID, 
      globalConf->kickoffPositionsOffensive[newPlayerNumber - 1] );
  setInitialPosition( BEFORE_KICK_OFF, THEIR_TEAM_ID, 
      globalConf->kickoffPositionsDefensive[newPlayerNumber - 1] );
  setInitialPosition( PLAY_ON, BOTH_TEAMS_ID, 
      globalConf->homePositions[newPlayerNumber - 1] );
  theBasePositions = theInitialPositions;
  setRole( globalConf->roles[newPlayerNumber-1] );
}

const ModelledPlayer& WorldModel::player( int playerId ) const
{
  if ( playerId == SELF_ID ) {
    playerId = id();
  }
  if ( playerId == UNDEF_PLAYER_ID ) {
    throwException( InvalidIdException("Id was UNDEF_PLAYER_ID in \
        WorldModel::player( int )") );
  } 
 
  return theSituations[currentSituationIndex].playerWithId( playerId );
}

const ModelledBall& WorldModel::ball() const
{
  return theSituations[currentSituationIndex].ball();
}

const Situation& WorldModel::situation( int timeStepOffset = 0 ) const
{
  if ( abs(timeStepOffset) > historyCount ) {
    throwException( Array<Situation>::IndexOutOfBoundsException( 
      "timeStep offset too large." ) );
  }
  int index = (currentSituationIndex + timeStepOffset + situationFieldSize) % 
      situationFieldSize;
  if ( index < 0 ) {
    throwException( Array<Situation>::IndexOutOfBoundsException( 
      "negative index is illegal" ) );
  }
    
  if ( timeStepOffset > 0 && 
       theSituations[currentSituationIndex].simStep() + timeStepOffset >
       theSituations[index].simStep() ) {
    theSituations[index] = theSituations[currentSituationIndex ];
    theSituations[index].simulateSteps( timeStepOffset );
  }
  return theSituations[index];
}
  
void WorldModel::setInitiative( int newInitiative ) 
{
  if ( newInitiative == OUR_TEAM_ID || newInitiative == THEIR_TEAM_ID ||
       newInitiative == BOTH_TEAMS_ID ) {
    theInitiative = newInitiative;
  }
  else if ( newInitiative == LEFT_SIDE_ID || newInitiative == RIGHT_SIDE_ID) {
    if ( side() == newInitiative ) {
      theInitiative = OUR_TEAM_ID; 
    }
    else {
      theInitiative = THEIR_TEAM_ID;
    }
  }
  else {
    theInitiative = NO_TEAM_ID;
  }
}

bool WorldModel::hasInitiative( int aTeamId ) const
{
  bool answer;
  switch ( aTeamId ) {
    case OUR_TEAM_ID:
      answer = ( initiative() == OUR_TEAM_ID || 
                 initiative() == BOTH_TEAMS_ID );
      break;
    case THEIR_TEAM_ID:
      answer = ( initiative() == THEIR_TEAM_ID || 
          initiative() == BOTH_TEAMS_ID );
      break;
    default:
      throwException( InvalidIdException( 
        "Only OUR_TEAM_ID and THEIR_TEAM_ID are allowed as parameters." ) );
  }
  return answer;
}

bool WorldModel::amIOneOfTheNearestPlayerToBall() const
{
  int stepsOfNearestPlayer = 
    theSituations[currentSituationIndex].stepsOfNearestPlayer( OUR_TEAM_ID );
  int stepsOfMyself =
    theSituations[currentSituationIndex].stepsOfPlayerWithId( id() );

  if( stepsOfMyself - stepsOfNearestPlayer <= 
      globalConf->interceptionTolerance ) {
    return YES;
  }
  else {
    return NO;
  }
}

int WorldModel::score( int aTeamId ) const
{
  switch ( aTeamId ) {
  case OUR_TEAM_ID:
    return theScore[0];
    break;
  case THEIR_TEAM_ID:
    return theScore[1];
    break;
  default:
    throwException( InvalidIdException( 
        "score for unknown team id demanded." ) );
  }
}

void WorldModel::setScore(int newScore, int aTeamId ) 
{
  switch ( aTeamId ) {
  case OUR_TEAM_ID:
    theScore[0] = newScore;
    break;
  case THEIR_TEAM_ID:
    theScore[1] = newScore;
    break;
  default:
    throwException( InvalidIdException( 
      "You try to set score for an unknown team id." ) );
  }
}

void WorldModel::integrateSensorInfo()
{
  Situation newSituation;
  bool isOldSituationCorrect = YES;

  INFO( Log::WORLD_MODEL, 2, "ENTER WorldModel::integrateSensorInfo\n" );
  INFO( Log::WORLD_MODEL, 9, "Situation before synchronize:\n" +
        situation().toString() + "\n" );

  try {
    //  alte Situationen aufbereiten.
    synchronizeSituations( theSensors->simStep() );
    INFO( Log::WORLD_MODEL, 9, "Situation after synchronize:\n" +
          situation().toString() + "\n" );
  } 
  catch( IncorrectOldSituationException ) {
    isOldSituationCorrect = NO;
    INFO( Log::WORLD_MODEL, 9, "Old Situation was incorrect\n" );
  }
  // neue Situation anlegen
  newSituation = Situation( theSensors->simStep() );

  // Daten des eigenen Spielers in der Situation festlegen
  newSituation.addPlayer( modelOfSelf() );

  // Daten der anderen Objekte in die neue Situation aufnehmen.
  addSeenObjects( newSituation );

  if( isOldSituationCorrect ) {
    INFO( Log::WORLD_MODEL, 9, "Situation before merge:\n" +
          newSituation.toString() + "\n" );
    // Erweitern der neuen Situation um Objekte der alten Situation.
    mergeSituations( newSituation, theSituations[currentSituationIndex] );
  }  

  // Hoerinformationen auswerten (Schiedrichter und andere Spieler)
  processAudioInfo( newSituation, theSensors->heardMessages() );
  // neue Situation fest speichern
  theSituations[currentSituationIndex] = newSituation;

  setDecisionReason( DECISION_REASON_SEE_INFO );

  // Konsistenzchecks auf der aktuellen Situation
  // Hier wird im Tournament-Mode mehr gemacht als normal
  improveConsistency();

  INFO( Log::WORLD_MODEL, 6, "New current Situation:\n" +
        theSituations[currentSituationIndex].toString() + "\n" );
  INFO( Log::SITUATION, 6, "OffsideLineX: " + String( 
      newSituation.offsideLineX() ) + "\n" );
  INFO( Log::WORLD_MODEL, 2, "LEAVE WorldModel::integrateSensorInfo\n" );
}

void WorldModel::integrateHearInfo()
{
  //  alte Situationen aufbereiten.
  synchronizeSituations( theSensors->simStepOfSoccerServer() );

  // Hoerinformationen auswerten (Schiedrichter und andere Spieler)
  processAudioInfo( theSituations[currentSituationIndex], 
                    theSensors->heardMessages() );

  setDecisionReason( DECISION_REASON_REFEREE_DECISION );  

  // Konsistenzchecks auf der aktuellen Situation
  improveConsistency();
}
  
void WorldModel::synchronizeSituations( int aSimStep )
{
  Situation newSituation;
  int stepDifference = aSimStep - simStep();

  INFO( Log::WORLD_MODEL, 4, "ENTER WorldModel::synchronizeSituations\n" );
  INFO( Log::WORLD_MODEL, 13, "gameState: " + String( gameState() ) + "\n" );
  
  if( fabs(stepDifference) > historyCount ) {
    throwException( IncorrectOldSituationException() );
  }

  if( stepDifference > 0 ) {
    // ich muss noch simulieren
    INFO( Log::WORLD_MODEL, 10, "I have to simulate " + 
          String( stepDifference ) + " steps.\n" );
    while( stepDifference > 0 ) {
      newSituation = situation();
      newSituation.simulateSteps( 1 );
      addCurrentSituation( newSituation );
      stepDifference--;
    }
  }
  else if( stepDifference < 0 ) {
    // aktuelle Situation wurde noch nicht vom SoccerServer erreicht
    if( gameState() != BEFORE_KICK_OFF &&
        gameState() != TIME_OVER &&
        gameState() != GOAL &&
        gameState() != OFFSIDE ) {
      // Zeit im SoccerServer ist weiter gelaufen, also Situationen 
      // zur"ucksetzen
      INFO( Log::WORLD_MODEL, 10, "I have to switch to an older situation (" + 
            String( -stepDifference ) + " steps back).\n" );
      while( stepDifference < 0 ) {
        removeCurrentSituation();
        stepDifference++;
      }
      if( situation().simStep() != aSimStep ) {
        throwException( Exception( "Situation field is corrupt" ) );
      }
    }
    else {
      // Zeit ist im SocccerServer nicht weiter gelaufen, Zeitindizes der
      // entsprechenden Situationen zur"ucksetzen
      INFO( Log::WORLD_MODEL, 10, "I set the time back (" + 
            String( -stepDifference ) + " steps).\n" );

      if( situation( stepDifference ).simStep() != aSimStep ) {
        throwException( Exception( 
            "prepareOldSituation: Situation field is corrupt" ) );
      }
      stepDifference++;
      int index;
      while( stepDifference <= 0 ) {
        index = (currentSituationIndex + stepDifference + situationFieldSize) %
          situationFieldSize;
        INFO( Log::WORLD_MODEL, 13, "Accessing situation[" +
              String( index ) + "]\n" );
        theSituations[index].setSimStep( aSimStep );
        stepDifference++;
      }
    }
  }
  else {
    // alles ok.
  }
  INFO( Log::WORLD_MODEL, 4, "LEAVE WorldModel::synchronizeSituations\n" );
}

const ModelledPlayer WorldModel::modelOfSelf() const
{
  // Spieler holen
  ModelledPlayer playerSelf = situation().playerWithId( id() );
  
  // Aktuellen Viewpoint bestimmen
  try {
    ViewPoint currentViewPoint = theField.triangulateViewPoint( 
        theSensors->sensedLines(), theSensors->sensedFlags() );

    if( theSide == LEFT_SIDE_ID ) {
      playerSelf.setPos( currentViewPoint.pos() );
      playerSelf.setFaceDir( currentViewPoint.faceDir() );
    }
    else if( theSide == RIGHT_SIDE_ID ) {
      Vector pos = Vector( -currentViewPoint.pos().x(),
                           -currentViewPoint.pos().y() );
      playerSelf.setPos( pos );
      playerSelf.setFaceDir( currentViewPoint.faceDir() - PI );
    }
    else {
      throwException( InvalidIdException( "Undef side Id" ) );
    }
    playerSelf.setPosReliability( Reliability::PERFECT );
    playerSelf.setFaceDirReliability( Reliability::PERFECT );
  }
  catch( Field::FailedPositionCalculationException ) {
    // FaceDir neu errechnen, sonst die alten Werte nehmen
    double currentFaceDir = 
      theField.computeFaceDir( theSensors->sensedLines() );

    if( theSide == LEFT_SIDE_ID ) {
      playerSelf.setFaceDir( currentFaceDir );
    }
    else if( theSide == RIGHT_SIDE_ID ) {
      playerSelf.setFaceDir( currentFaceDir - PI );
    }
    else {
      throwException( InvalidIdException( "Undef side Id" ) );
    }
    playerSelf.setFaceDirReliability( Reliability::PERFECT );
  }
  catch( Field::FailedTriangulationException ) {
    // Alten ViewPoint "ubernehmen
  }
  
  // Geschwindigkeit, Stamina und Effort aus bodyInfos auslesen
  BodyInfos bodyInfos = theSensors->lastBodyInfos();
  if( bodyInfos.time() == simStep() ) {
    // nur auswerten, wenn aktuelle BodyInfos vorhanden.
    // Speed auswerten; hier k"onnte man auch auf "altere
    // BodyInfos zur"uckgreifen und simulieren
    playerSelf.setSpeed( Vector( bodyInfos.amountOfSpeed(), 
                                 playerSelf.speed().angle(),
                                 POLAR ) );
    playerSelf.setSpeedReliability( Reliability::PERFECT );
    
    // VisibleAngle, Stamina und Effort setzen
    playerSelf.setViewWidth( bodyInfos.viewWidth() );
    playerSelf.setStamina( bodyInfos.stamina() );
    playerSelf.setEffort( bodyInfos.effort() );
    playerSelf.setStaminaReliability( Reliability::PERFECT );
    
    // hier eventuell weitere Werte aus BodyInfos auswerten
  }

  return playerSelf;
}

void WorldModel::addSeenObjects( Situation& newSituation )
{
  ModelledPlayer myself = newSituation.playerWithId( id() );
  Reliability posReliability = Reliability( max( 
      myself.faceDirReliability().get(), myself.posReliability().get() ) );
  Reliability speedReliability = Reliability( max( 
      myself.faceDirReliability().get(), myself.speedReliability().get() ) );

  // Ball hinzufuegen
  ModelledBall newBall = ModelledBall( theSensors->sensedBall() );
  if ( !newBall.isUndef() ) {
    if ( newBall.pos().isValid() ) {
      // Blickwinkel wegrechnen
      Vector relativBallPos = relativePos( newBall.pos(), myself.faceDir() );
      // Koordinatentransformation in das absolute System
      newBall.setPos( myself.pos() + relativBallPos );
      newBall.setPosReliability( posReliability );
      
      if ( newBall.speed().isValid() ) {
        newBall.setSpeed( myself.speed() + 
            relativeSpeed( relativBallPos, newBall.speed() ) );
        newBall.setSpeedReliability( speedReliability );
      }
    }
  }
  else {
    newBall = ModelledBall( BALL_ID, Vector(0, 0), Vector(0, 0) );
  }
  newSituation.setBall( newBall );
  
  // Der Reihe nach alle gesehenen Spieler hinzufuegen
  ModelledPlayer newPlayer;
  int playerCount = theSensors->sensedPlayers().count();
  for (int i = 0; i < playerCount; i++) {
    newPlayer = ModelledPlayer( theSensors->sensedPlayers()[i] );
    if( newPlayer.pos().isValid() ) {
      // Blickwinkel wegrechnen
      Vector relativePlayerPos = relativePos( newPlayer.pos(),
                                              myself.faceDir() );
      newPlayer.setPos( myself.pos() + relativePlayerPos );
      newPlayer.setPosReliability( posReliability );

      if ( newPlayer.speed().isValid() ) {
	newPlayer.setSpeed( myself.speed() +
            relativeSpeed( relativePlayerPos, newPlayer.speed() ) );
	newPlayer.setSpeedReliability( speedReliability );
      } 
      else {
	newPlayer.setSpeed( Vector(0, 0) );
      } 

      if ( newPlayer.faceDir() != UNDEF_DOUBLE ) {  
	newPlayer.setFaceDir( newPlayer.faceDir() + myself.faceDir() );
        newPlayer.setFaceDirReliability( myself.faceDirReliability() );
      }
      else {
	newPlayer.setFaceDir( 0 );
      }
    }
    else {
      newPlayer.setPos( Vector( 0, 0 ) );
      newPlayer.setSpeed( Vector( 0, 0 ) );
      newPlayer.setFaceDir( 0 );
      WARNING( Log::WORLD_MODEL, "See a player without a position." );
    }
    newSituation.addPlayer( newPlayer );
  }
}

ModelledBall WorldModel::mergeBalls( const ModelledBall& trustedBall, 
                             const ModelledBall& anotherBall, 
                             const ModelledPlayer& myself ) const
{
  ModelledBall thisBall = trustedBall;

  if( anotherBall.posReliability().get() < Reliability::RANDOM && 
      anotherBall.posReliability().get() < 
      thisBall.posReliability().get() ) {
    if( !myself.isInFieldOfSense( anotherBall ) ) {
      thisBall.setId( anotherBall.id() );
      thisBall.setPos( anotherBall.pos() );
      thisBall.setPosReliability( anotherBall.posReliability() );
    }
  }
  if( anotherBall.speedReliability().get() < Reliability::RANDOM && 
      anotherBall.speedReliability().get() < 
      thisBall.speedReliability().get() ) {
    thisBall.setSpeed( anotherBall.speed() );
    thisBall.setSpeedReliability( anotherBall.speedReliability() );
  }
  if( anotherBall.lastTouchedAtTime() > thisBall.lastTouchedAtTime() ) {
    thisBall.setLastTouchedBy( anotherBall.lastTouchedBy() );
    thisBall.setLastTouchedAtTime( anotherBall.lastTouchedAtTime() );
  }  
  return thisBall;
}

ModelledPlayer WorldModel::mergePlayers( const ModelledPlayer& trustedPlayer, 
                               const ModelledPlayer& anotherPlayer ) const 
{
  ModelledPlayer thisPlayer = trustedPlayer;

  if( anotherPlayer.posReliability().get() < Reliability::RANDOM && 
      anotherPlayer.posReliability().get() < 
      thisPlayer.posReliability().get() ) {
    thisPlayer.setPos( anotherPlayer.pos() );
    thisPlayer.setPosReliability( anotherPlayer.posReliability() );
  }
  if( anotherPlayer.speedReliability().get() < Reliability::RANDOM && 
      anotherPlayer.speedReliability().get() < 
      thisPlayer.speedReliability().get() ) {
    thisPlayer.setSpeed( anotherPlayer.speed() );
    thisPlayer.setSpeedReliability( anotherPlayer.speedReliability() );
  }
  if( anotherPlayer.faceDirReliability().get() < Reliability::RANDOM && 
      anotherPlayer.faceDirReliability().get() < 
      thisPlayer.faceDirReliability().get() ) {
    thisPlayer.setFaceDir( anotherPlayer.faceDir() );
    thisPlayer.setFaceDirReliability( anotherPlayer.faceDirReliability() );
  }
  if( anotherPlayer.staminaReliability().get() < Reliability::RANDOM && 
      anotherPlayer.staminaReliability().get() < 
      thisPlayer.staminaReliability().get() ) {
    thisPlayer.setViewWidth( anotherPlayer.viewWidth() );
    thisPlayer.setStamina( anotherPlayer.stamina() );
    thisPlayer.setEffort( anotherPlayer.effort() );
    thisPlayer.setRecovery( anotherPlayer.recovery() );
    thisPlayer.setStaminaReliability( anotherPlayer.staminaReliability() );
  }
  if( !thisPlayer.isPlayerNumberKnown() ) {
    thisPlayer.setId( anotherPlayer.id() );
  }
  
  return thisPlayer;
}

void WorldModel::mergeSituations( Situation& trustedSituation,
                                  Situation& anotherSituation ) const
{
  INFO( Log::WORLD_MODEL, 4, "ENTER WorldModel::mergeSituations\n" );
  if( trustedSituation.simStep() != anotherSituation.simStep() ) {
    throwException( 
        Exception( "Can only merge situations with the same simStep." ) );
  }
  ModelledPlayer myself = trustedSituation.playerWithId( id() );

  // Ball mischen
  INFO( Log::WORLD_MODEL, 10, "Merging ball: \ntrustedBall: " +
        trustedSituation.ball().toString() + "\nanotherBall: " +
         anotherSituation.ball().toString() + "\n" );
  trustedSituation.setBall( mergeBalls( trustedSituation.ball(),
      anotherSituation.ball(), myself ) );
  INFO( Log::WORLD_MODEL, 10, "Result of merge, ball:\n" +
        trustedSituation.ball().toString() + "\n" );

  // Spieler mischen
  int anotherPlayerCount = anotherSituation.playerCount();
  int trustedPlayerCount = trustedSituation.playerCount();
  bool isMerged;
  double minDist;
  int nearestTrustedPlayerIndex;

  for( int i = 0; i < anotherPlayerCount; i++ ) {
    const ModelledPlayer& anotherPlayer = 
      anotherSituation.playerWithIndex( i );
    if( anotherPlayer.isPlayerNumberKnown() ) {
      isMerged = NO;
      minDist = globalConf->maximalMergeDist;
      nearestTrustedPlayerIndex = UNDEF_INT;
      for( int j = 0; j < trustedPlayerCount; j++ ) {
        const ModelledPlayer& trustedPlayer = 
          trustedSituation.playerWithIndex( j );
        if( trustedPlayer.isPlayerNumberKnown() ) { // Spielernummer bekannt
          if( anotherPlayer.id() == trustedPlayer.id() ) {
            INFO( Log::WORLD_MODEL, 8, "Merge player index: " + String( i ) +
                  " with index: " + String( j ) + "\n" );
            INFO( Log::WORLD_MODEL, 10, "Merging player with same known " +
                  String( "number: \ntrustedPlayer: " ) +
                  trustedPlayer.toString() + "\nanotherPlayer; " +
                  anotherPlayer.toString() + "\n" );
            trustedSituation.setPlayer( mergePlayers( trustedPlayer,
                anotherPlayer ), j );
            INFO( Log::WORLD_MODEL, 10, "Result of merge, player:\n" +
                  trustedSituation.playerWithIndex( j ).toString() + "\n" );
            isMerged = YES;
            break;
          }
        }
        else { // muss auf einen der Spieler im Blickfeld abgebildet werden
          if( myself.isInFieldOfSense( anotherPlayer ) ) {
            double dist = anotherPlayer.distTo( trustedPlayer.pos() );
            if( dist < minDist ) {
              minDist = dist;
              nearestTrustedPlayerIndex = j;
            }
          }
        }
      }
      if( !isMerged ) {
        if( !myself.isInFieldOfSense( anotherPlayer ) ) { // uebernehmen
          INFO( Log::WORLD_MODEL, 8, "Add player index: " + String(i) + "\n" );
          trustedSituation.addPlayer( anotherPlayer );
          INFO( Log::WORLD_MODEL, 10, "Result of add, player:\n" +
                anotherPlayer.toString() + "\n" );
        }
        else { // abbilden auf Spieler, der am naehsten steht
          if( nearestTrustedPlayerIndex != UNDEF_INT ) {
            INFO( Log::WORLD_MODEL, 8, "Merge player index: " + String(i) + 
                  " with index: " + String( nearestTrustedPlayerIndex ) +
                  "\n" );
            INFO( Log::WORLD_MODEL, 10, "Merging player without knowning " +
                  String( "the number of trustedPlayer: \ntrustedPlayer: " ) + 
                  trustedSituation.playerWithIndex( 
                      nearestTrustedPlayerIndex ).toString() +
                  "\nanotherPlayer; " + anotherPlayer.toString() + "\n" );
            trustedSituation.setPlayer( mergePlayers( 
                trustedSituation.playerWithIndex( nearestTrustedPlayerIndex ),
                anotherPlayer ), nearestTrustedPlayerIndex );
            INFO( Log::WORLD_MODEL, 10, "Result of merge, player:\n" +
                  trustedSituation.playerWithIndex( 
                      nearestTrustedPlayerIndex ).toString() + "\n" );
          }
          else { // an den Rand verschieben?
            Vector vectorToPlayer = anotherPlayer.pos() - myself.pos();
            vectorToPlayer.setPolar( vectorToPlayer.length(), 
                signum( normalizeAngle( vectorToPlayer.angle() ) ) *
                myself.visibleAngle() / 2 );
            Vector newPosOfPlayer = myself.pos() + vectorToPlayer;
            if( anotherPlayer.distTo( newPosOfPlayer ) <= 
                globalConf->escapeFieldOfViewDist ) {
              ModelledPlayer modifiedPlayer( anotherPlayer );
              modifiedPlayer.setPos( newPosOfPlayer );
              trustedSituation.addPlayer( modifiedPlayer );
            }
            else { // zu dem Spieler gibt es keinen der passt.
              WARNING( Log::WORLD_MODEL, "Can't merge player with index " +
                  String( i ) + " to a trusted player and dist is to " +
                  "large. Will forget him." );
            }
          }
        }
      }
    }
  }


  INFO( Log::WORLD_MODEL, 4, "LEAVE WorldModel::mergeSituations\n" );

}

int WorldModel::indexOfCorrespondingPlayer( ModelledPlayer player,
                                           Situation situation ) const 
{
  ModelledPlayer anotherPlayer;
  for( int i = 0; i < situation.playerCount(); i++ ) {
    anotherPlayer = situation.playerWithIndex( i );
    if( ( anotherPlayer.id() > MIN_OPPONENT_ID && 
          anotherPlayer.id() < MAX_EXISTING_OPPONENT_ID ) ||
        ( anotherPlayer.id() > MIN_TEAMMATE_ID && 
          anotherPlayer.id() < MAX_EXISTING_TEAMMATE_ID ) ) {
      if( anotherPlayer.id() == player.id() ) {
        return i;
      }
      else {
        // player hat keine genaue id, aber vielleicht stimmt er ja doch
        // mit diesem Spieler "uberein
        // erstmal noch offen, vielleicht sp"ater mehr

      }
    }
  }

  throwException( NoCorrespondingPlayerException() );  
}

void WorldModel::processAudioInfo( Situation& newSituation, 
				   const Array<Message>& messages ) {
  int i;
  int messageCount = messages.count();
  for(i = 0; i < messageCount; i++) {
    switch ( messages[i].senderInfo() ) {
      case REFEREE_ID:
        // Eine Schiedsrichterentscheidung. Die ist wichtig.
        processRefereeMessage( messages[i] );
        break;
      case SELF_ID:
        // eigene Nachricht wurde tatsaechlich vom Server versendet.
        // Interessant. Wird aber momentan nicht ausgewertet.
        break;
      case THEIR_TEAM_ID:
        // hier koennte man fremde Nachrichten speichern, um sie spaeter
        // zur Verwirrung abzuspielen.
        break;
      case OUR_TEAM_ID:
        processOurTeamMessage( newSituation, messages[i] );
        break;
      default:
        WARNING( Log::WORLD_MODEL, "Unidentified Message-Sender. This " +
                 String( "should not happen." ) );
    }
  }
}


String WorldModel::encode( int number ){

  int element, offset;
  const int maximalExponent = 3;
  char encodedValue;
  String string;

  for( int i = maximalExponent; i >= 0; i-- ) {
    element = (int) ( number / pow( 36, i ) );
    if( (element == 0) && ( string.length() == 0 ) && ( i > 0 ) ) {
      continue;
    }
    
    if( element > 9 ) {
      offset = 65-10;
    }
    else { 
      offset = 48;
    }
    sprintf( &encodedValue, "%c", element + offset );
    number -= ( element * (int) ( pow( 36, i ) ) );
    string = string + String( &encodedValue );
  }

  return string;
}



int WorldModel::decode( String string ){
  int decodedPartValue;
  int decodedValue = 0;
  char message[MAX_MESSAGE];
  char encodedValue;
  strcpy( message, string.cString() );

  for( int i = string.length() - 1; i >= 0; i-- ) {
    sscanf( message, "%c %[^\n]", &encodedValue, message );
    decodedPartValue = (int) encodedValue;
    if (decodedPartValue <= 57) {
      decodedPartValue -= 48;
    }
    else {
      decodedPartValue -= 65-10;
    }
    decodedPartValue = decodedPartValue * (int) ( pow( 36, i ) );
    decodedValue += decodedPartValue;
  }

  return decodedValue;
}



void WorldModel::processOurTeamMessage( Situation& newSituation,
                                        const Message& message )
{
  INFO( Log::WORLD_MODEL, 4, "ENTER WorldModel::processOurTeamMessage\n" );
  
  char simStep[3];
  char kindOfSubString;
  String theMessage = message.message();  
  char newMessage[MAX_MESSAGE];
  double id, stamina, playerPosX, playerPosY, ballPosX, ballPosY,
    ballSpeedX, ballSpeedY;
  unsigned long ballPosReliability, ballSpeedReliability, staminaReliability, 
    playerPosReliability;
    
  ModelledPlayer player = ModelledPlayer( UNDEF_ID, Vector( 0, 0 ),
					  Vector( 0, 0 ), 0 );   
  ModelledBall ball = ModelledBall( BALL_ID, Vector( 0, 0 ), 
				      Vector( 0, 0 ) );
  Situation mySituation = Situation();

  sscanf( theMessage.cString(), "%s %[^\n]", simStep, newMessage );
  mySituation.setSimStep(globalWorldModel->decode( simStep ));


  int stepDifference = newSituation.simStep() - mySituation.simStep();
   
  INFO( Log::WORLD_MODEL, 11, "stepDifference: " + String( stepDifference ) +
        "\n" );
  if( stepDifference > 2 ){
    return;
  }  
  
  sscanf( newMessage, "%c %[^\n]", &kindOfSubString, newMessage );
  while( kindOfSubString != '+') {
    switch( kindOfSubString ) {
    case 'p':
      sscanf( newMessage, "%lf %lf %ld %[^\n]", &ballPosX, &ballPosY, 
	      &ballPosReliability, newMessage);
      ball.setPos( Vector( ballPosX, ballPosY ) );
      ball.setPosReliability( ballPosReliability );
      break;
    case 's':
      sscanf( newMessage, "%lf %lf %ld %[^\n]", &ballSpeedX, 
	      &ballSpeedY, &ballSpeedReliability, newMessage );
      ball.setSpeed( Vector( ballSpeedX, ballSpeedY ) );
      ball.setSpeedReliability( ballSpeedReliability );
      break;
    default:
      WARNING( Log::WORLD_MODEL, "Suspect say-Message." );
    }
    sscanf( newMessage, "%c %[^\n]", &kindOfSubString, newMessage );
  }
  mySituation.setBall( ball );
  
  while( kindOfSubString == '+' ) { 
    ModelledPlayer newPlayer = player;
    sscanf( newMessage, "%lf %[^\n]", &id, newMessage);
    newPlayer.setId( id );
    sscanf( newMessage, "%c %[^\n]", &kindOfSubString, newMessage );
    
    while( kindOfSubString != '+' && kindOfSubString != '-' ){
      switch( kindOfSubString ){
        case 'o': 
	        sscanf( newMessage, "%lf %lf %ld %[^\n]", &playerPosX, &playerPosY,
		          &playerPosReliability, newMessage );
	        newPlayer.setPos( Vector( playerPosX, playerPosY ) );
	        newPlayer.setPosReliability( playerPosReliability );
	        sscanf( newMessage, "%c %[^\n]", &kindOfSubString, newMessage );
	        break;   
        case 't':
	        sscanf( newMessage, "%lf %ld %[^\n]",
		          &stamina, &staminaReliability, newMessage ); 
	        newPlayer.setStamina( stamina );
	        newPlayer.setStaminaReliability( staminaReliability );
	        sscanf( newMessage, "%c %[^\n]", &kindOfSubString, newMessage );
	        break;
        default:
	        WARNING( Log::WORLD_MODEL, "Suspect say-Message." );
	        sscanf( newMessage, "%c %[^\n]", &kindOfSubString, newMessage );
      } 
    }
    mySituation.addPlayer( newPlayer );
  }
  
  INFO( Log::WORLD_MODEL, 13 , "gehoerte Situation:\n" +
        mySituation.toString() + "\n\n" );
  
  if( stepDifference > 0 ) {
    mySituation.simulateSteps( stepDifference );
  }
  
  INFO( Log::WORLD_MODEL, 8 , "aufgearbeitete Situation:\n" +
        mySituation.toString() + "\n" );
  
  if( stepDifference >= 0 ) {
    mergeSituations( newSituation, mySituation );
  }
  
  INFO( Log::WORLD_MODEL, 4, "LEAVE WorldModel::processOurTeamMessage\n" );
}


void WorldModel::processRefereeMessage( const Message& message ) 
{
  setDecisionReason( DECISION_REASON_REFEREE_DECISION );
  setSimStepOfLastRefereeDecision( simStep() );
  ModelledBall ball;
  
  String messageString = message.message();
  if ( messageString == "before_kick_off" ) {
    setGameState( BEFORE_KICK_OFF );
    // Links faengt an.
    setInitiative( LEFT_SIDE_ID ); 
  } 
  else if ( messageString == "time_over" ) {
    setGameState( TIME_OVER );
    setInitiative( NO_TEAM_ID );
  } 
  else if ( messageString == "play_on" ) {
    setGameState( PLAY_ON );
    setInitiative( BOTH_TEAMS_ID );
  }
  else if ( messageString == "kick_off_l" ) {
    setGameState( KICK_OFF );
    setInitiative( LEFT_SIDE_ID ); 
  }
  else if ( messageString == "kick_off_r" ) {
    setGameState( KICK_OFF );
    setInitiative( RIGHT_SIDE_ID ); 
  } 
  else if ( messageString == "kick_in_l" ) {
    setGameState( KICK_IN );
    setInitiative( LEFT_SIDE_ID ); 
  }
  else if ( messageString == "kick_in_r" ) {
    setGameState( KICK_IN );
    setInitiative( RIGHT_SIDE_ID ); 
  }
  else if ( messageString == "free_kick_l" ) {
    setGameState( FREE_KICK );
    setInitiative( LEFT_SIDE_ID );
    ball = theSituations[currentSituationIndex].ball();
    ball.setPosReliability( Reliability::RANDOM );
    theSituations[currentSituationIndex].setBall( ball );
  }
  else if ( messageString == "free_kick_r" ) {
    setGameState( FREE_KICK );
    setInitiative( RIGHT_SIDE_ID ); 
    ball = theSituations[currentSituationIndex].ball();
    ball.setPosReliability( Reliability::RANDOM );
    theSituations[currentSituationIndex].setBall( ball );
  }
  else if ( messageString == "drop_ball" ) {
    setGameState( DROP_BALL );
    setInitiative( BOTH_TEAMS_ID ); 
  }  
  else if ( messageString == "corner_kick_l" ) {
    setGameState( CORNER_KICK );
    setInitiative( LEFT_SIDE_ID ); 
  }
  else if ( messageString == "corner_kick_r" ) {
    setGameState( CORNER_KICK );
    setInitiative( RIGHT_SIDE_ID ); 
  }
  else if ( messageString == "goal_kick_l" ) {
    setGameState( GOAL_KICK );
    setInitiative( LEFT_SIDE_ID ); 
  }
  else if ( messageString == "goal_kick_r" ) {
    setGameState( GOAL_KICK );
    setInitiative( RIGHT_SIDE_ID ); 
  }
  else if ( messageString == "offside_l" ) {
    setGameState( OFFSIDE );
    setInitiative( RIGHT_SIDE_ID ); 
  }
  else if ( messageString == "offside_r" ) {
    setGameState( OFFSIDE );
    setInitiative( LEFT_SIDE_ID ); 
  }
  else if ( messageString.beginsWith( "goal_l" ) ) {
    int newScore = messageString.subString( 8 ).asInt();
    setGameState( GOAL );
    if (side() == LEFT_SIDE_ID ) {
      setScore( newScore, OUR_TEAM_ID );
    }
    else {
      setScore( newScore, THEIR_TEAM_ID );
    }
    setInitiative( RIGHT_SIDE_ID ); 
  }
  else if ( messageString.beginsWith( "goal_r" ) ) {
    int newScore = messageString.subString( 8 ).asInt();
    setGameState( GOAL );
    if (side() == RIGHT_SIDE_ID ) {
      setScore( newScore, OUR_TEAM_ID );
    }
    else {
      setScore( score( THEIR_TEAM_ID ) + 1, THEIR_TEAM_ID );
    }
    setInitiative( LEFT_SIDE_ID ); 
  }
  else if ( messageString.beginsWith( "foul_l" ) ) {
    setGameState( FOUL );
    setInitiative( RIGHT_SIDE_ID ); 
  }
  else if ( messageString.beginsWith( "foul_r" ) ) {
    setGameState( FOUL );
    setInitiative( LEFT_SIDE_ID ); 
  }
  else if ( messageString == "half_time" ) {
    INFO( Log::WORLD_MODEL, 8, "Halbzeit: Setze Stamina-Modell zurueck.\n" );
    setGameState( HALF_TIME );
    setInitiative( NO_TEAM_ID );
    ModelledPlayer myself = player();
    myself.setStaminaValues( globalConf->playerStaminaMax, 1.0, 1.0,
        Reliability::PERFECT );
    theSituations[currentSituationIndex].setPlayer(  myself );
  }
  else if ( messageString == "time_up" ) {
    setGameState( TIME_UP );
    setInitiative( NO_TEAM_ID ); 
  }
  else if ( messageString == "time_up_without_a_team" ) {
    setGameState( TIME_UP_WITHOUT_A_TEAM );
    setInitiative( NO_TEAM_ID ); 
  }
  else if ( messageString == "time_extended" ) {
    setGameState( TIME_EXTENDED );
    setInitiative( NO_TEAM_ID ); 
  }
  else if ( messageString.beginsWith( "goalie_catch_ball" ) ) {
    // Nichts machen. Gleich darauf kommt ein Freekick. Dann wird 
    // die Initiative richtig gesetzt.
  }
  else {
    WARNING( Log::WORLD_MODEL, "Unknown referee decision '" +
             messageString + "' ignored." );
  }
}

void WorldModel::addCurrentSituation( const Situation& aSituation )
{
  currentSituationIndex = ( currentSituationIndex + 1 + situationFieldSize) % 
      situationFieldSize;
  theSituations[currentSituationIndex] = aSituation;
}

void WorldModel::removeCurrentSituation()
{
  theSituations[currentSituationIndex].setSimStep( 0 );
  currentSituationIndex = ( currentSituationIndex - 1 + situationFieldSize) % 
      situationFieldSize;
}

Vector WorldModel::relativePos( const Vector& seenPos, double faceDir )
{
  Vector relativePos;
  relativePos.setPolar( seenPos.length(), seenPos.angle() + faceDir );
  return relativePos;
}
  
Vector WorldModel::relativeSpeed( const Vector& relativePos,
				  const Vector& relativeChangeOfPos )
{
  Vector nextPos;
  if( relativeChangeOfPos.wasInitializedWithNegativeDist() ) {
    nextPos.setPolar( relativePos.length() - relativeChangeOfPos.length(),
                      relativePos.angle() + relativeChangeOfPos.angle() - PI );
  }
  else {
    nextPos.setPolar( relativePos.length() + relativeChangeOfPos.length(),
                      relativePos.angle() + relativeChangeOfPos.angle() );
  }
  Vector relativeSpeed = nextPos - relativePos;
  return relativeSpeed;
}


void WorldModel::updateWithAction( const Action& anAction ) 
{
  INFO( Log::WORLD_MODEL, 2, "ENTER WorldModel::updateWithAction\n" );
  INFO( Log::WORLD_MODEL, 9, "Updating WorldModel with action" +
      anAction.toString() + "\n" );
  Situation newSituation = situation();
  
  ModelledPlayer myself = newSituation.playerWithId( id() );
  ModelledBall ball = newSituation.ball();

  switch( anAction.id() ) {
    case DASH_ACTION:
      myself.accelerate( anAction.firstParameter() );
      break;
    case TURN_ACTION:
      myself.turn( anAction.firstParameter() );
      break;
    case KICK_ACTION:
      try {
        ball.accelerate( anAction.firstParameter(), 
            anAction.secondParameter(), myself );
      }
      catch ( ModelledBall::KickImpossibleException ) {
        // Ball kann nicht gekickt werden
      }
      ball.setLastTouchedBy( myself.id() );
      ball.setLastTouchedAtTime( newSituation.simStep() );
      break;
    case MOVE_ACTION:
      myself.setPos( Vector( anAction.firstParameter(), 
			     anAction.secondParameter() ) );
      break;
    case CATCH_ACTION:
      setSimStepOfLastCatch( simStep() );      
      break;
    case CHANGE_VIEW_ACTION:
      myself.setViewWidth( anAction.viewWidth() );
      theSituations[currentSituationIndex].setPlayer( myself );
      break;
    case SAY_ACTION:
      setSimStepOfLastSay( simStep() );
      break;
    case SENSE_BODY_ACTION:
      break;
    case EMPTY_ACTION:
      break;
    default:
      throwException( Action::ActionException(
          "Unknown ActionId in WorldModel::updateWithAction." ) );
      break;
  }
  
  if( anAction.isMainAction() ) {
    newSituation.setBall( ball );  
    newSituation.setPlayer( myself );
    newSituation.simulateSteps( 1 );
  
    addCurrentSituation( newSituation );
    INFO( Log::WORLD_MODEL, 11, "New situation is " + situation().toString() +
          "\n" );
  }
  INFO( Log::WORLD_MODEL, 2, "LEAVE WorldModel::updateWithAction\n" );
}

void WorldModel::registerExecution( const Action& anAction ) 
{
  INFO( Log::WORLD_MODEL, 2, "ENTER WorldModel::registerExecution\n" );
  if ( anAction.id() != SENSE_BODY_ACTION && 
       anAction.id() != SAY_ACTION ) {
    if( gameState() != BEFORE_KICK_OFF &&
        gameState() != TIME_OVER &&
        gameState() != GOAL &&
        gameState() != OFFSIDE ) {
      theExecutedActions.addObject( anAction );
      INFO( Log::WORLD_MODEL, 8, "Registered action: " + anAction.toString() +
            "\n" );
    }
    updateWithAction( anAction );
  }
  INFO( Log::WORLD_MODEL, 2, "LEAVE WorldModel::registerExecution\n" );
}  

  

const double WorldModel::OUTER_LIMIT = 28.0;
const double WorldModel::INNER_LIMIT = 20.0;

int WorldModel::region( const Vector& pos ) const
{
  const double MAIN_DIVIDER = 15;
  const double WING_DISTRIBUTOR_DIVIDER = -20;
  const double WING_DISTRIBUTOR_INNER_LIMIT = 7.1;
  double x = pos.x();
  double y = pos.y();
  if ( x < MAIN_DIVIDER ) {
    // Teil des Spielfelds, der die Mittellinie und das eigene Tor
    // enthaelt
    if ( y > OUTER_LIMIT ) {
      // oben, Feldrand
      return 6;
    }
    if ( y > INNER_LIMIT ) {
      // oben, Streifen darunter
      return 4;
    }
    if ( y < -OUTER_LIMIT ) {
      // unten, Feldrand
      return 7;
    }
    if ( y < -INNER_LIMIT ) {
      // unten, darueber
      return 5;
    }
    if ( x < WING_DISTRIBUTOR_DIVIDER ) {
      if ( y > WING_DISTRIBUTOR_INNER_LIMIT ) {
        // Verteiler auf den oberen Fluegel
        return 1;
      }
      if ( y < -WING_DISTRIBUTOR_INNER_LIMIT ) {
        // Verteiler auf den unteren Fluegel
        return 2;
      }
    }
    // Der Rest ist eine T-foermige Region mit dem Querbalken quer zur
    //Mittellinie und dem Laengsbalken zum eigenen Tor hin
    return 3;
  }
  else {
    // Teil des Spielfelds, das das gegnerische Tor enthaelt
    double m;
    double c;
    // Steigung(m) und Achsenabschnitt(c) einer Geraden berechnen, die
    // vom oberen Torpfosten zum Punkt (18.5, 34) geht
    m = ( theField.goalWidth()/2 - theField.fieldWidth()/2 ) / 
      ( theField.fieldLength()/2 - 18.5 );
    c = theField.fieldWidth()/2 - 18.5 * m;
    if ( y > m * x + c || y > OUTER_LIMIT ) {
      // y liegt oberhalb der Geraden oder oberhalb von OUTER_LIMIT
      return 9;
    }
    // Steigung(m) und Achsenabschnitt(c) einer Geraden berechnen, die
    // vom unteren Torpfosten zum Punkt (18.5, -34) geht
    m = ( -theField.goalWidth()/2 + theField.fieldWidth()/2 ) / 
      ( theField.fieldLength()/2 - 18.5 );
    c = -theField.fieldWidth()/2 - 18.5 * m;
    if ( y < m * x + c || y < -OUTER_LIMIT ) {
      // y liegt unterhalb der Geraden oder unterhalb von OUTER_LIMIT
      return 10;
    }
    // Das uebrigbleibende Rechteck mit zwei abgeschnittenen Ecken. Die kurze
    // Seite ist das gegnerische Tor
    return 8;
  } 
}


Vector WorldModel::preferredKickDirection( const Vector& pos ) const
{
  INFO( Log::WORLD_MODEL, 3, "ENTER WorldModel::preferredKickDirection\n" );
  Vector toTarget;
  const Vector WING_ATTRACTOR( 0, 34 );
  Vector opponentAttractor;
  if ( player().pos().y() > 0 ) {
    opponentAttractor.set( 52.5, 4 );
  }
  else {
    opponentAttractor.set( 52.5, -4 );
  }
      
  
  const Vector CENTER_ATTRACTOR( 0, 0);
  const double maximumSpeed = globalConf->ballSpeedMax;
  const double slowSpeed = maximumSpeed / 5;
  // Die Vorzugsrichtung ist abhaengig von der Region.
  int regionOfPos = region( pos );
  INFO( Log::WORLD_MODEL, 8, "I am in region " + String( regionOfPos ) +
        "\n" );
  switch ( regionOfPos ) {
    case 1:
      // Raus auf den oberen Fluegel
      toTarget = WING_ATTRACTOR - pos;
      toTarget.setPolar( maximumSpeed, toTarget.angle() );
      break;
    case 2:
      // Raus auf den unteren Fluegel
      toTarget = ( -1.0 * WING_ATTRACTOR ) - pos;
      toTarget.setPolar( maximumSpeed, toTarget.angle() );
      break;
    case 3:
    case 4:
    case 5:
      // waagrecht rueber in die gegnerische Haelfte
      toTarget = opponentAttractor;
      toTarget.setPolar( maximumSpeed, toTarget.angle() );
      break;
    case 6:
    case 7:
      // weg vom Rand, direkt auf das Tor des Gegners
      toTarget = opponentAttractor - pos;
      toTarget.setPolar( slowSpeed, toTarget.angle() );
      break;
    case 8:
      // direkt aufs gegnerische Tor
      toTarget = opponentAttractor - pos;
      toTarget.setPolar( maximumSpeed, toTarget.angle() );
      break;
    case 9:
    case 10:
      toTarget = Vector( opponentAttractor.x() - 11, opponentAttractor.y() ) -
          pos;
      toTarget.setPolar( maximumSpeed, toTarget.angle() );
      break;
    default:
        throwException( InvalidIdException( "Invalid region number." ) );
  }
  INFO( Log::WORLD_MODEL, 6, "Preferred kick direction is " + 
        toTarget.toString() + "\n" );
  return toTarget;
}

int WorldModel::role() const
{
  return theRole;    
}

void WorldModel::setRole( int newRole )
{
  theRole = newRole;
}

int WorldModel::positionIndexFromState( int aGameState, int anInitiative ) const
{
  int index;
  if ( aGameState == UNDEF_INT ) {
    aGameState = globalWorldModel->gameState();
  }
  if ( anInitiative == UNDEF_INT ) {
    anInitiative = globalWorldModel->initiative();
  }
  
  switch ( aGameState ) {
    case WorldModel::KICK_OFF:
    case WorldModel::BEFORE_KICK_OFF:
    case WorldModel::GOAL:
      if ( anInitiative == OUR_TEAM_ID ) {
        index = 0;
      } 
      else {
        index = 1;
      }
      break;
    case WorldModel::PLAY_ON:
    default:
      index = 2;
  }
  
  return index;
}


String WorldModel::toString() const
{
  String output;
  output = String( "WorldModel Info: simStep: " + String( simStep() ) +
                   "\n  gameState: " + String( gameState() ) +
                   "\n  initiative: " + String( initiative() ) +
                   "\n  simStepOfLastCatch: " + String( simStepOfLastCatch() ) +
                   "\n  decisionReson: " + String( decisionReason() ) +
                   "\n  current Situation: " + situation().toString() +
                   "  Base Positions: (" );
  for( int i = 0; i < theBasePositions.count(); i++) {
    output = output + "\n   " + theBasePositions[i].toString();
  }
  output = output + " )\n\n";

  return output;
}


void WorldModel::setBasePosition( 
    int aGameState, int anInitiative, const Vector& pos )
{
  theBasePositions[positionIndexFromState( aGameState, anInitiative)] = pos;
}

void WorldModel::setInitialPosition( 
    int aGameState, int anInitiative, const Vector& pos )
{
  theInitialPositions[positionIndexFromState( aGameState, anInitiative)] = pos;
}

Vector WorldModel::basePosition( int aGameState, int anInitiative ) const
{
  return theBasePositions[positionIndexFromState( aGameState, anInitiative) ];
}

Vector WorldModel::initialPosition( int aGameState, int anInitiative ) const
{
  return theInitialPositions[positionIndexFromState( 
      aGameState, anInitiative) ];
}


void WorldModel::improveConsistency()
{
  // Bereits fuer diesen simStep abgeschickte Aktionen
  // draufsimulieren
  Array<Action> simulatedActions;
  int oldSituationSimStep = simStep();
  
  INFO( Log::WORLD_MODEL, 8, "Checking the reexecution of " +
      String( theExecutedActions.count() ) + " actions\n" );

  for( int i = 0; i < theExecutedActions.count(); i++ ) {
    if( theExecutedActions[i].simStep() == 
        theSituations[currentSituationIndex].simStep() ) {
      INFO( Log::WORLD_MODEL, 2, "Reexecuting " +
            theExecutedActions[i].toString() + " \n" );

      if( decisionReason() == DECISION_REASON_SEE_INFO &&
	  theExecutedActions[i].id() == TURN_ACTION ) {
        // turns wurden bereits im SoccerServer eingearbeitet
        theExecutedActions[i] = Action( EMPTY_ACTION, 
            theExecutedActions[i].simStep() );
      }

      ModelledPlayer playerSelf = 
        theSituations[currentSituationIndex].playerWithId( id() );
      double oldRecovery = playerSelf.recovery();

      updateWithAction( theExecutedActions[i] );
      simulatedActions.addObject( theExecutedActions[i] );

      if( decisionReason() == DECISION_REASON_SEE_INFO &&
	  theExecutedActions[i].id() == DASH_ACTION ) {
        const BodyInfos& bodyInfos = theSensors->lastBodyInfos();
        if( theExecutedActions[i].actionCount() < bodyInfos.dashs() ) {
          // dash wurde schon im senseBody eingerechnet
          INFO( Log::WORLD_MODEL, 8, "New senseBody after dash -> " +
                String( "set new stamina values from senseBody.\n" ) );
          INFO( Log::WORLD_MODEL, 11, "Dash count before send of dash: " +
                String( theExecutedActions[i].actionCount() ) +
                "   Dash count at senseBody: " + String( bodyInfos.dashs() ) +
                + "\nreread bodyInfo: " + bodyInfos.toString() + "\n" );

          playerSelf.setViewWidth( bodyInfos.viewWidth() );
          playerSelf.setStamina( bodyInfos.stamina() );
          playerSelf.setEffort( bodyInfos.effort() );
          playerSelf.setRecovery( oldRecovery );
          playerSelf.setStaminaReliability( Reliability::PERFECT );
          playerSelf.recover();

          theSituations[currentSituationIndex].setPlayer( playerSelf );
        }
      }

    } 
  }
  theExecutedActions.removeAll();
  theExecutedActions.addObjects( simulatedActions );

  INFO( Log::WORLD_MODEL, 8, "new simStep: " + String( simStep() ) +
        "\told simStep: " + String( oldSituationSimStep ) + "\n" );
  
  if( simStep() - oldSituationSimStep > 1 ) {
    WARNING( Log::WORLD_MODEL, "action for a future situation was sended" );
  }  

  // Ggf. gameState auf PLAY_ON stellen. Dadurch koennen Spieler
  // nachgestartet werden.
  // Im Testmodus ist dieser Check abgeschaltet, damit man zwischendurch
  // das Spiel anhalten kann, indem man auf BEFORE_KICK_OFF schaltet.
  if ( globalConf->tournamentMode == YES ) {
    if ( ( gameState() == BEFORE_KICK_OFF ) && 
         ( situation().simStep() % globalConf->halfTime != 0 ) ) {
      setGameState( PLAY_ON );
      setInitiative( BOTH_TEAMS_ID );
    }
  }
}


