#ifndef _AGENT_H_
#define _AGENT_H_

#include "Communicator.h"
#include "Deliberator.h"
#include "Skills.h"
#include "Sensors.h"
#include "Effectors.h"
#include "Timer.h"
#include "Exception.h"

/**
  Die zentrale Hauptklasse. 
  Eine Instanz wird beim Programmstart erzeugt.
  Agent steuert die Initialisierung aller Komponenten und startet 
  die Hauptschleife.

  Von der Klasse wird nur ein Objekt erzeugt. Die Klasse ist nicht zur
  Vererbung geeignet.  

  {\small {\bf Quelltext}

  [\URL[Header]{src/Agent.h} | [\URL[Implementierung]{src/Agent.cc}]

  {\bf CVS Informationen zum Interface}  
  \begin{verbatim}
  **************************************************
  *
  * Aktueller Stand
  *
  *   $RCSfile: Agent.h,v $
  *   $Revision: 1.20 $
  *   $Date: 1998/08/14 13:19:54 $
  *   $Author: myritz $ 
  *
  **************************************************
  \end{verbatim} }*/
class Agent
{
public:

  /** Fehler, der bei nicht gueltiger oder fehlgeschlagener
    Initialisierung auftritt. */
  class InitFailedException: public Exception 
    {
    public:
      InitFailedException(const String& aDescription )
      	: Exception( aDescription ) {}
    };
  
  /**Der einzige Konstruktor.
    Er startet alle notwendigen Komponenten ueber die internen
    Init-Funktionen.
    @param argv Paare -<Option> <Wert> ; -h liefert Hilfe. */
  Agent (int argc, char* argv []);
  
  /**Der Destruktor.
    Er deinstalliert alle Komponenten und raeumt auf. Logdateien werden
    geschlossen. */
  ~Agent ();
  
  /**Startet den Ueberlegungsprozess. 
    Das Ausfuehren der Handlungen und das Aufnehmen neuer Sensorinformationen
    geschieht asynchron ueber den Scheduler. */
  void run ();
  
  
private:
  /** Der UDP/IP-level Kommunikationsverwalter */
  Communicator *theCommunicator;
  
  /** Verwaltet die Sensorik */
  Sensors* theSensors;
  
  /** Die agierende Komponente */
  Effectors* theEffectors;
  
  /** Die Ueberlegungskomponente. Frueher Brain und Entscheidung. */
  Deliberator* theDeliberator;
  
  String teamName;
  int playerNumber;
  bool isGoalie;
  String serverName;
  int serverPort;
  String serverConfigurationFile;
  String agentConfigurationFile;
  Array<String> extendedConfigurationFiles;
  bool isHelpWanted;

  /** Die Parameter werden eingelesen. */
  void readParameter( int argCount, char* argv [] );  

  /** Die Effektoren werden initialisiert und gestartet. */
  void startComponents();

  /** Initialisiert den Zufallszahlengenerator. */
  void initRandomNumberGenerator();
  
  /** Schreibt Hilfeinformationen nach stdout.
    @param programName Name des aufgerufenen Programmes. */
  void showHelp(String programName);
  
};

#endif
