#ifndef _EFFECTORS_H_
#define _EFFECTORS_H_

#include <stdio.h>

#include "Communicator.h"
#include "String.h"
#include "Action.h"
#include "Plan.h"
#include "Exception.h"
#include "BodyInfos.h"

/**
  Handlungsausfuehrende Komponente. 
  Die Basisaktionen, die der Server zur Verfuegung stellt, sind hier gekapselt.
  Effectors uebernimmt ausserdem das Queueing der Kommandos und das
  Benachrichtigen des Weltmodells ueber tatsaechlich abgeschickte Aktionen.

  Alle Aktionen, zu denen es Action-Objekte gibt, koennen mit queueAction 
  ausgefuehrt werden. Ausserdem koennen Plaene, also Abfolgen von Aktionen mit
  queuePlan in die Warteschlange gehaengt werden.

  Die Aktionen enterField, reEnterField und leaveField muessen direkt
  ueber den Aufruf der gleichnamigen Methode ausgeloest werden.

  Von der Klasse wird nur ein Objekt erzeugt. 
  Die Klasse ist nicht zur Vererbung geeignet.

  {\small {\bf Quelltext}

  [\URL[Header]{src/Effectors.h} | [\URL[Implementierung]{src/Effectors.cc}]

  {\bf CVS Informationen zum Interface}
  \begin{verbatim}
  **************************************************
  *
  * Aktueller Stand
  *
  *   $RCSfile: Effectors.h,v $
  *   $Revision: 1.22 $
  *   $Date: 1998/08/14 13:20:27 $
  *   $Author: myritz $ 
  *
  **************************************************
  \end{verbatim} }*/
class Effectors
{
public:
  /** Interne Fehlerklasse. */
  class EffectorsException: public Exception 
  {
  public:
    EffectorsException(const String& aDescription )
      : Exception( aDescription ) {}
  };
  
  /** Interne Fehlerklasse. Wird aufgerufen, wenn beim Wunsch Kommandos
    zu senden keine Kommandos mehr in der Queue sind. */
  class NothingToSendException: public Exception
  {
  public:
    NothingToSendException() :  Exception() {}
  };
    
  /** Erzeugt ein neues Effectors-Objekt. 
    @param aCommunicator Der Communicator, ueber den Effectors die
    Aktionen verschickt. */
  Effectors( Communicator* aCommunicator );
  
  /** Betritt das Spielfeld zur ersten Halbzeit auf der Seite des Teams
    aTeamName. Ist das nicht moeglich, wird eine EffectorsException erzeugt. Im
    Weltmodell werden eingetragen:
    \begin{itemize}
    \item Die Spielernummer, die der Server beim Anmelden zurueckschickt
    \item Die Seite, auf der der Spieler spielt
    \item Der Teamname
    \end{itemize}
    Das Weltmodell muss also vor Aufruf dieser Methode bereits initialisiert
    sein. */
  void enterField( const String& aTeamName, bool goalie = NO );

  /** Betritt das Spielfeld zur zweiten Halbzeit auf der Seite des Teams
    aTeamName. Ist das nicht moeglich, wird eine EffectorsException erzeugt. Im
    Weltmodell werden eingetragen:
    
    \begin{itemize}
    \item Die bereits bekannte Spielernummer,
    \item Die Seite, auf der der Spieler spielt
    \item Der Teamname.
    \end{itemize}
    Das Weltmodell muss also vor Aufruf dieser Methode bereits initialisiert
    sein. */
  void reEnterField( const String& aTeamName, int aPlayerNumber );
	
  /** Veranlasst das Verlassen des Feldes. */
  void leaveField();
  
  /** Haengt eine einzelne Aktion hinten an die Warteschlange der
    auszufuehrenden Aktionen. */
  void queueAction( const Action& anAction );
	
  /** Fuegt alle Aktionen eines Plans hinten an die Warteschlange der
    auszufuehrenden Aktionen an. */
  void queuePlan( const Plan& aPlan );
  
  /** Loescht alle in der Queue befindlichen Aktionen. */
  void clearQueue();
  
  /** Die Warteschlange der auszufuehrenden Aktionen wird gesetzt. Dazu
    wird der alte Plan geloescht und der neue angefuegt. Diese Methode 
    sollte wenn moeglich den Methoden queuePlan und queueAction 
    vorgezogen werden. */
  void setPlan( const Plan& aPlan );

  /** Fuehrt eine Aktion sofort aus, ohne sie in die Warteschlange zu haengen.
    Die Erstellung des korrekten Strings fuer
    den Server liegt in der Verantwortung der Aktion. Es findet keinerlei
    Fehlerueberpruefung des Strings statt. Die Aktion wird allerdings vorher
    gefragt, ob sie meint, einen gueltigen CommandString erzeugen zu koennen.
    Wenn ja, wird er gesendet, wenn nicht, wird eine EffectorsException
    ausgeloest.  */
  void performAction( const Action& anAction );
  
  /** 
    Fuehrt die naechste Hauptaktion (mainAction) des gegenwaertigen Planes
    aus. Die eventuell vor der Hauptaktion liegenden Nebenaktionen 
    (subActions) werden verworfen. Die Hauptaktion wird mittels 
    performAction abgeschickt.
    @exception NothingToSendException wenn es keine Hauptaktion mehr im
      Plan gibt. */
  void performMainAction( const BodyInfos& aBodyInfo );

  
  /** 
    Fuehrt die naechste Aktion, des gegenwaertigen Planes aus, wenn diese
    eine Nebenaktion (subAction) ist. Ist die naechste Aktion eine
    Hauptaktion (MainAction) wird eine NothingToSendException erzeugt. Die 
    Nebenaktion wird mittels performAction abgeschickt.
    @exception NothingToSendException wenn es bis zur n"achsten Hauptkation
      oder bis zum Ende des Planes keine Nebenaktion mehr gibt. */
  void performSubAction();
	
  /**
    Meldet ob sich im Plan noch eine Aktion befindet. Wenn
    dies der Fall ist wird YES zurueckgegeben , ansonsten NO. */
  bool isActionLeft();

  /** Beschreibung als String fuer die Ausgabe auf einem Terminal. */
  String toString() const;  
	
private:
  /** Pointer auf den Communicator des Agenten. */
  Communicator *theCommunicator;
  
  /** Ist das Feld betreten? Die Menge der erlaubten Aktionen ist davon
    abhaengig. */
  bool hasEnteredField;
  
  /** Die Aktionsqueue. */
  Array<Action> theQueue;
  
  /** Der Index der gegenwaertig aktuellen Aktion. */
  int indexOfCurrentAction;
  
};

#endif

