
#ifndef _RASTEREDPLAYGROUND_H_
#define _RASTEREDPLAYGROUND_H_


/**
  Modelliert ein gerastertes Spielfeld zum Finden von guten passiven 
  Positionen. Jeder Spieler will als Passivaktion eine gute passive Position 
  finden und sich zu dieser begeben. Dazu lernt er in regelm\"a\3igen 
  Abst\"anden die Schwerpunkte der gegnerischen Spieler (wenn der Ball weit 
  entfernt ist) und den Ballverlauf (wenn der Ball nah ist).

  Die Utility ist \"aquivalent zur Entfernung des Balles, d. h. je weiter der 
  Ball entfernt ist, desto gr\"o\3er ist die N\"utzlichkeit, eine gute passive
  Position zu erreichen.

 {\bf CVS Informationen}
 \begin{verbatim}
 **************************************************
 *
 * Aktueller Stand
 *
 *   $RCSfile: RasteredPlayground.h,v $
 *   $Revision: 1.3 $
 *   $Date: 1998/08/14 13:21:08 $
 *   $Author: myritz $ 
 *
 **************************************************
 \end{verbatim}
 */

#include "Colouring.h"
#include "Vector.h"
#include "Array.h"

class RasteredPlayground {
public:
  /** Der Konstruktor.
   Initialisiert das gerasterte Spielfeld. */
  RasteredPlayground();

  /** Der Destruktor. */
  ~RasteredPlayground();
  
  /// Spielfeldkoord. (reelle Koord.)
  struct PositionTyp { 
    Vector thePosition;
    int ballAppearance;
    int edgeColour; 
  };

  /** Berechnet eine gute passive Position.
   @returns Position */
  Vector goodPassivePosition();

  /** Lernt (erfasst Daten ueber) die Spielaufstellung 
    der Gegner und des Balles. */
  void acquireKnowledgeAbout(const int about);
  
private:
  /// Eine Spalte von Colouring Objekten
  typedef Array < Colouring > ColumnT;
  
  /// Das 2D-Array von Farben. Repraesentiert das Spielfeld gerastert.
  Array < ColumnT > colours;

  /// Faerbt das Spieler-Feld auf Position pos (die vom Weltmodell).
  void incColourOfPlayerAt(Vector& pos);

  /// Faerbt das Ball-Feld auf Position pos (die vom Weltmodell).
  void incColourOfBallAt(Vector& pos);

  /// Zahl der Zeilen im Raster
  int rows;
  /// Zahl der Spalten im Raster
  int columns;

  /// Haeufiger benutzter Wert: Die halbe Spielfeldlaenge
  double halfFieldLength;
  /// Haeufiger benutzter Wert: Die halbe Spielfeldbreite
  double halfFieldWidth;

  /// Umrechnung von Rasterkoordinaten in kartesische Spielfeldkoordinaten
  Vector realFromRasteredCoordinates(const Vector& rasteredVector);

  /** Umrechnung von kartesischen Spielfeldkoordinaten in Rasterkoordinaten */
  Vector rasteredFromRealCoordinates(const Vector& realVector);

  /** Berechnung einer Position auf der Kante 
      zwischen den Punkten "firstVector" und "secondVector" */
  Vector positionOnEdge( Vector firstVector, 
      Vector secondVector );

  /** Prueft ob ein anderer Mitspieler naeher an der moeglichen
      passiven Position ist. */
  bool teammateCloserTo( const Vector &thePositionOnEdge, 
      Array  <Vector> &teammatePos );
  
  /** Berechnung der Schwerpunkte der Gegnerpositionen 
      @return Zahl der Schwerpunkte */
  int foundFocuses(Array <Vector> & playerFocus);

  /** Auswahl von nicht zu kurzen und nicht zu langen Kanten
      zwischen den Schwerpunkten,
      zu denen kein anderer Spieler naeher steht
      und Berechnung von Positionen auf diesen Kanten
      @return YES = Positionen gefunden, NO = Positionen nicht gefunden */
  bool computePositions(Array <PositionTyp> & position, 
      Array <Vector> & playerFocus );

  /** Berechnung der besten Position nach den Kriterien
      Abstand von home-Position, Abstand von aktueller Position, 
      Faerbung  der Schwerpunkte der Kante, Ballverlauf */
  Vector bestPosition(Array <PositionTyp>& position );

  /** Speichert die Positionen der eigenen Mitspieler. 
      Die Positionen werden aus dem Weltmodell geholt. 
      @param teammatePos Das Array, in das die Positionen der Mitspieler
      eingetragen werden. */
  void storePosOfTeammates( Array <Vector>& teammatePos );
 
  /** Berechnung der Ballhaeufigkeit 
      (= Summe der Ballfaerbungen in den Rastern) entlang einer Kante */
  int getBallAppearance( Vector firstVector,
      Vector secondVector );
  
};

#endif
